package com.searchtechnologies.aspire.standalone;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.function.Consumer;

import org.w3c.dom.Element;

import com.searchtechnologies.aspire.components.FilesystemRAP;
import com.searchtechnologies.aspire.connector.framework.controller.CrawlControllerImpl;
import com.searchtechnologies.aspire.connector.framework.fetcher.FetchURL;
import com.searchtechnologies.aspire.connector.services.GroupListener;
import com.searchtechnologies.aspire.connector.services.RepositoryAccessProvider;
import com.searchtechnologies.aspire.connector.services.RepositoryAccessProvider.GroupExpansion;
import com.searchtechnologies.aspire.connector.services.RepositoryAccessProvider.RepositoryConnection;
import com.searchtechnologies.aspire.connector.services.ScanListener;
import com.searchtechnologies.aspire.connector.services.SourceInfo;
import com.searchtechnologies.aspire.connector.services.SourceItem;
import com.searchtechnologies.aspire.framework.ComponentImpl;
import com.searchtechnologies.aspire.framework.JobFactory;
import com.searchtechnologies.aspire.services.AspireException;
import com.searchtechnologies.aspire.services.AspireObject;
import com.searchtechnologies.aspire.services.Component;
import com.searchtechnologies.aspire.services.Job;
import com.searchtechnologies.aspire.services.groupexpansion.UoG;
import com.searchtechnologies.aspire.services.noSql.NoSQLConnection;

/**
 * Standalone java program for running an Aspire Connector without Aspire
 * @author aaguilar
 *
 */
public class Main {
  
  public static final String RAP = "rap";
  static {
    try {
      AspireObject ao = new AspireObject("config");
      ao.add("rap", RAP);
      emptyDom = ao.toDom();
    } catch (AspireException e) {
      emptyDom = null;
    }
  }
  static Element emptyDom;

  public static void main(String[] args) throws Exception {

    //Get the scan properties from the arguments
    String scanPropertiesFile = args[0];
    
    //Instantiate a new RepositoryAccessProvider from the connector implementation
    JobFactory.initializeForStandAloneUsage();
    
    ComponentImpl component = new FilesystemRAP();
    component.initialize(emptyDom);
    RepositoryAccessProvider rap = (RepositoryAccessProvider) component;
    
    //Create a CrawlControllerImpl, as some connectors depends on it
    StandaloneCrawlController crawlCtrlImpl = new StandaloneCrawlController(rap);
    
    //Load the content-source.xml configuration into an AspireObject
    AspireObject scanProps = new AspireObject("doc");
    AspireObject scanPropsFile = AspireObject.createFromXML(new File(scanPropertiesFile));
    scanProps.add(scanPropsFile);
    
    //Create and initialize a new SourceInfo from the RAP
    SourceInfo info = rap.newSourceInfo(scanPropsFile);
    info.initialize(scanProps);
    info.setCrawlController(crawlCtrlImpl);
    crawlCtrlImpl.setSourceInfo(info);
    
    //execute the crawl
    crawl(rap, info, Main::downloadStream);
    
    //Execute the users and groups download process
    downloadGroups(rap, info);
    
    //Close any opened resources by the components
    component.close();
    crawlCtrlImpl.close();
  }
  
  /**
   * Receives the items from the connector and write the stream into a file
   * @param item
   */
  private static void downloadStream(SourceItem item) {
    System.out.println("Item: "+item.getName());
    try {
      InputStream is = item.getContentStream();
      if (is != null) {
        FileOutputStream fos = new FileOutputStream(new File("output/"+item.getName()));
        copyStream(is, fos);
        fos.close();
        is.close();
      }
    } catch (Exception e) {
      e.printStackTrace();
    }
  }
  
  private static void downloadGroups(RepositoryAccessProvider rap,
      SourceInfo info) throws AspireException {
    GroupExpansion ge = rap.getGroupExpansion(info);
    if (ge != null) {
      ArrayList<UoG> emptyList = new ArrayList<>();
      RepositoryConnection conn = rap.newConnection(info);
      HashMap<String, ArrayList<UoG>> membership = new HashMap<String, ArrayList<UoG>>();
      GroupListener groupListener = new GroupListener() {
        @Override
        public void addMemberOf(UoG userOrGroup, UoG group) throws AspireException {
          if (!membership.containsValue(userOrGroup.toString())){
            membership.put(userOrGroup.toString(), new ArrayList<UoG>());
          }
          membership.get(userOrGroup.toString()).add(group);
          System.out.println("UserOrGroup: "+userOrGroup +", group: "+group.toString());
        }
        
        @Override
        public void addMemberOf(UoG userOrGroup, Collection<UoG> groups)
            throws AspireException {
          if (!membership.containsValue(userOrGroup.toString())){
            membership.put(userOrGroup.toString(), new ArrayList<UoG>());
          }
          membership.get(userOrGroup.toString()).addAll(groups);
          System.out.println("UserOrGroup: "+userOrGroup +", groups: ");
          for (UoG group : groups) {
            System.out.println("-- "+group.toString());
          }
        }
        
        @Override
        public void addMember(UoG group, UoG member) throws AspireException {
          addMemberOf(member,group);
        }
        
        @Override
        public void addMember(UoG group, Collection<UoG> members)
            throws AspireException {
          for (UoG userOrGroup : members) {
            addMemberOf(userOrGroup, group);
          }
        }
      };
      ge.downloadUsersAndGroups(conn, groupListener , emptyList);
      conn.closeConnection();
    }
  }
  
  private static void copyStream(InputStream input, OutputStream output) throws IOException {
    byte[] buffer = new byte[1024]; // Adjust if you want
    int bytesRead;
    while ((bytesRead = input.read(buffer)) != -1) {
      output.write(buffer, 0, bytesRead);
    }
  }
  
  private static void crawl(RepositoryAccessProvider rap, SourceInfo info, Consumer<SourceItem> processor) throws AspireException {
    //The ScanListener which maintain the local processQueue and listen for new items to crawl
    Scanner scanner = new Scanner(rap, info);
    
    //Create the crawlRoot item to initialize the crawl from the RAP
    SourceItem crawlRoot = new SourceItem("crawlRoot");
    rap.processCrawlRoot(crawlRoot, info, scanner);
    
    //Crawls the local processQueue while it is not empty
    // when empty, it means the crawl finished
    do {
      scanner.processQueue(processor);
    } while (!scanner.isQueueEmpty());
  }

  static class Scanner implements ScanListener {

    /**
     * The queue that receives all the new items discovered
     */
    ArrayList<SourceItem> queue;
    
    /**
     * Temporary queue used for iterate the original queue
     */
    ArrayList<SourceItem> safeQueue;
    private RepositoryAccessProvider rap;
    private SourceInfo info;
    private FetchURL fetcher;
    
    public Scanner(RepositoryAccessProvider rap, SourceInfo info) throws AspireException {
      this.rap = rap;
      this.info = info;
      queue = new ArrayList<SourceItem>();
      safeQueue = new ArrayList<SourceItem>();
      fetcher = new StandaloneFetchURL(rap);
      System.out.println(fetcher.getStringFromConfig(emptyDom, "rap", null));
      fetcher.initialize(emptyDom);
    }
    
    public void close() {
      fetcher.close();
    }

    @Override
    public void addItem(SourceItem item) {
      //This gets called when the RAP scanner adds an item to crawl
      queue.add(item);
    }

    @Override
    public void addItems(List<SourceItem> items) {
      //This gets called when the RAP scanner adds items to crawl
      queue.addAll(items);
    }
    
    public void processQueue(Consumer<SourceItem> processor) throws AspireException {
      
      RepositoryConnection conn = rap.newConnection(info);
      
      //Move the items from the original queue into the safeQueue
      //And clear the orignal
      safeQueue.clear();
      safeQueue.addAll(queue);
      queue.clear();
      
      for (SourceItem item : safeQueue) {
        boolean container = false;
        
        if (rap.isContainer(item, conn)) {
          container = true;
        }
        
        if (info.indexContainers() || !container) {
          rap.populate(item, info, conn);
          //Call the fetcher
          Job job = JobFactory.newInstance(item.generateJobDocument());
          job.put("sourceInfo", info);
          job.put("crawlController", info.getCrawlController());
          fetcher.process(job);
          item.setContentStream((InputStream) job.get("contentStream"));
        }
        
        if (container) {
          rap.scan(item, info, conn, this);
        }
        processor.accept(item);
      }
      safeQueue.clear();
    }
    
    public boolean isQueueEmpty() {
      return queue.size()+safeQueue.size() == 0;
    }
  }
  
  private static class StandaloneFetchURL extends FetchURL {

    RepositoryAccessProvider rap;
    
    public StandaloneFetchURL(RepositoryAccessProvider rap) {
      this.rap = rap;
    }
    
    @Override
    public Component getComponent(String path) {
      if (RAP.equals(path)) {
        return (Component)rap;
      }
      return null;
    }
  } 
  
  private static class StandaloneCrawlController extends CrawlControllerImpl {

    RepositoryAccessProvider rap;
    public StandaloneCrawlController(RepositoryAccessProvider rap) {
      this.rap = rap;
    }

    @Override
    public RepositoryAccessProvider getRAP() {
      return rap;
    }

    @Override
    public NoSQLConnection getNoSQLConnection(String name, AspireObject properties) {
      return new EmptyNoSQLConnection();
    }
  }

}
